/* Replacement nl_types.h file for building GNU Emacs on Windows.
   Copyright (C) 2006, 2007 Free Software Foundation, Inc.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.  */

#ifndef _NL_TYPES_H
#define _NL_TYPES_H

typedef int nl_item;

#endif	/* _NL_TYPES_H */

/* arch-tag: 518aa765-b846-4a7a-a862-95fd022928a1
   (do not change this comment) */
