;;; sc.el --- old name for supercite

;; Maintainer: FSF

;; This file is part of GNU Emacs.

;;; Commentary:

;;; Code:

(require 'supercite)
(provide 'sc)

(message "The name `sc' works but is obsolete; please use `supercite' instead")

;;; arch-tag: 31e8ae19-689e-4b7d-9161-6d7dd60c6ece
;;; sc.el ends here
