/* Handle uxpds  */

#include "usg5-4-2.h"

/* This triggers a conditional in xfaces.c.  */
#define XOS_NEEDS_TIME_H

#define FSCALE 256

/* arch-tag: c02db2a6-c725-4a6b-adee-66e0303df2a8
   (do not change this comment) */
