;; -*- no-byte-compile: t -*-
(defun terminal-init-apollo ()
  "Terminal initialization function for apollo."  
  (load "term/vt100" nil t))

;;; arch-tag: c72f446f-e6b7-4749-90a4-bd68632adacf
;;; apollo.el ends here
